/*
 * Version 0.3
 *
 * Copyright 1999 by Red Hat GmbH, Deutschland. <Harald.Hoyer@redhat.de>
 *                                                                            
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is  hereby granted without fee, provided that
 * the  above copyright   notice appear  in   all  copies and  that both  that
 * copyright  notice   and   this  permission   notice  appear  in  supporting
 * documentation,  and   that   the  name  of  Red Hat GmbH  not  be  used  in
 * advertising or publicity pertaining to distribution of the software without
 * specific,    written    prior    permission.     Red Hat GmbH  makes     no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.                   
 *                                                                            
 * RED HAT GMBH DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT  SHALL RED HAT GMBH BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA  OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 *******************************************************************************
 *******************************************************************************
 *
 * This driver is able to deal with B112 touch panel.
 * It cannot initialise the touch panel to report absolute coordinates
 *
 *******************************************************************************
 *******************************************************************************
 */

#include "Xos.h"
#include <signal.h>
#include <stdio.h>

#define	 NEED_EVENTS
#include "X.h"
#include "Xproto.h"
#include "inputstr.h"
#include "scrnintstr.h"
#include "XI.h"
#include "XIproto.h"

#if defined(sun) && !defined(i386)
#include <errno.h>
#include <termio.h>
#include <fcntl.h>
#include <ctype.h>

#include "extio.h"
#else
#include "compiler.h"

#ifdef XFree86LOADER
#include "xf86_libc.h"
#endif
#include "xf86.h"
#include "xf86Procs.h"
#include "xf86_OSlib.h"
#include "xf86_Config.h"
#include "xf86Xinput.h"
#include "xf86Version.h"
#endif

#if !defined(sun) || defined(i386)
#include "os.h"
#include "osdep.h"
#include "exevents.h"

#include "extnsionst.h"
#include "extinit.h"
#endif

#if !defined(sun) || defined(i386)
/*
 ***************************************************************************
 *
 * Configuration descriptor.
 *
 ***************************************************************************
 */

#define PORT			1
#define B112_DEVICE_NAME	2
#define SCREEN_NO		3
#define UNTOUCH_DELAY		4
#define REPORT_DELAY		5
#define MAXX			6
#define MAXY			7
#define MIDX			8
#define MIDY			9
#define MINX			10
#define MINY			11
#define B112_EMULATE3        	12
#define B112_EM3TIMEOUT        	13
#define DEBUG_LEVEL     	14
#define HISTORY_SIZE		15
#define LINK_SPEED		16
#define ALWAYS_CORE		17

static SymTabRec B112Tab[] = {
  { ENDSUBSECTION,     "endsubsection" },
  { PORT,              "port" },
  { B112_DEVICE_NAME,   "devicename" },
  { SCREEN_NO,	       "screenno" },
  { UNTOUCH_DELAY,     "untouchdelay" },
  { REPORT_DELAY,      "reportdelay"},
  { MAXX,              "maximumxposition" },
  { MAXY,              "maximumyposition" },
  { MIDX,              "middlexposition" },
  { MIDY,              "middleyposition" },
  { MINX,              "minimumxposition" },
  { MINY,              "minimumyposition" },
  { B112_EMULATE3,	"emulate3buttons" },
  { B112_EM3TIMEOUT,	"emulate3timeout" },
  { DEBUG_LEVEL,       "debuglevel" },
  { HISTORY_SIZE,      "historysize" },
  { ALWAYS_CORE,       "alwayscore" },
  { -1,                "" },
};

static CARD32 buttonTimer(
#if NeedFunctionPrototypes
			  OsTimerPtr timer,
			  CARD32 now,
			  pointer arg
#endif
			  );
#endif
/*
 ***************************************************************************
 *
 * Default constants.
 *
 ***************************************************************************
 */

#define DEFAULT_MAX_X		985
#define DEFAULT_MID_X		507
#define DEFAULT_MIN_X		43
#define DEFAULT_MAX_Y		888
#define DEFAULT_MID_Y		466
#define DEFAULT_MIN_Y		62


/*
 ***************************************************************************
 *
 * Protocol constants.
 *
 ***************************************************************************
 */
#define B112_PACKET_SIZE		3
#define B112_BUFSIZE                    (B112_PACKET_SIZE*500)
#define B112_PORT "/dev/psaux"
#define B112_RESET "/var/tmp/resetB112"

/*
 ***************************************************************************
 *
 * Usefull macros.
 *
 ***************************************************************************
 */
#define WORD_ASSEMBLY(byte1, byte2)	(((byte2) << 8) | (byte1))
#define SYSCALL(call)	while(((call) == -1) && (errno == EINTR))
#define GET_TIME(tv) (gettimeofday(&tv, (struct timezone *)NULL))
#define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000+ \
                           (t2.tv_usec-t1.tv_usec)/1000)

#ifdef DBG
#undef DBG
#endif
#ifdef DEBUG
#undef DEBUG
#endif

static int      debug_level = 0;
#define DEBUG 1
#if DEBUG
#define DBG(lvl, f) {if ((lvl) <= debug_level) f;}
#else
#define DBG(lvl, f)
#endif


/*
 ***************************************************************************
 *
 * Device private records.
 *
 ***************************************************************************
 */
typedef struct _B112PrivateRec {
  char		*input_dev;	/* The touchscreen input tty		*/
  int		min_x;		/* Minimum x reported by calibration	*/
  int		mid_x;		/* Middle x				*/
  int		rmid_x;		/* Middle x				*/
  int		max_x;		/* Maximum x				*/
  int		min_y;		/* Minimum y reported by calibration	*/
  int		mid_y;		/* Middle y				*/
  int		rmid_y;		/* Middle y				*/
  int		max_y;		/* Maximum y				*/
  int		untouch_delay;	/* Delay before reporting an untouch (in ms) */
  int		report_delay;	/* Delay between touch report packets	*/
  int		screen_no;	/* Screen associated with the device	*/
  int		screen_width;	/* Width of the associated X screen	*/
  int		screen_height;	/* Height of the screen			*/
  Bool		inited;		/* The controller has already been configured?*/
  Bool		was_pressed;	/* The controller has already been configured?*/
  int           cur_x, cur_y;   /* current positione (last)             */
  int           avgx, avgy;     /* average over time, for smooth feeling*/
  int           upx, upy;       /* keep track of last finger-up place   */
  struct timeval lasttv;    /* time of events                       */
  OsTimerPtr    button_timer;   /* middle button emulation              */
  int		truebuttons;    /* buttons to press                    */
  int		lastButtons;    /* buttons to press                    */
  int		emulateState;    /* buttons to press                    */
  Bool          emulate3Buttons;
  int           emulate3Timeout;
  int		packet_buf_p;	/* Assembly buffer pointer		*/
  unsigned char	packet_buf[B112_BUFSIZE]; /* Assembly buffer		*/
} B112PrivateRec, *B112PrivatePtr;


#if !defined(sun) || defined(i386)


/*
 ***************************************************************************
 *
 * xf86B112Config --
 *	Configure the driver from the configuration data.
 *
 ***************************************************************************
 */
static Bool
xf86B112Config(LocalDevicePtr    *array,
	       int               inx,
	       int               max,
	       LexPtr            val)
{
  LocalDevicePtr        local = array[inx];
  B112PrivatePtr         priv = (B112PrivatePtr)(local->private);
  int                   token;

  while ((token = xf86GetToken(B112Tab)) != ENDSUBSECTION) {
    switch(token) {
      
    case PORT:
      if (xf86GetToken(NULL) != STRING)
	xf86ConfigError("B112 input port expected");
      priv->input_dev = strdup(val->str);	
      if (xf86Verbose)
	ErrorF("%s B112 input port: %s\n",
	       XCONFIG_GIVEN, priv->input_dev);
      break;

    case B112_DEVICE_NAME:
      if (xf86GetToken(NULL) != STRING)
	xf86ConfigError("B112 device name expected");
      local->name = strdup(val->str);
      if (xf86Verbose)
	ErrorF("%s B112 X device name: %s\n",
	       XCONFIG_GIVEN, local->name);      
      break;

    case SCREEN_NO:
      if (xf86GetToken(NULL) != NUMBER)
	xf86ConfigError("B112 screen number expected");
      priv->screen_no = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 associated screen: %d\n",
	       XCONFIG_GIVEN, priv->screen_no);      
      break;

    case UNTOUCH_DELAY:
      if (xf86GetToken(NULL) != NUMBER)
	xf86ConfigError("B112 untouch delay expected");
      priv->untouch_delay = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 untouch delay: %d ms\n",
	       XCONFIG_GIVEN, priv->untouch_delay*10);      
      break;

    case REPORT_DELAY:
      if (xf86GetToken(NULL) != NUMBER)
	xf86ConfigError("B112 report delay expected");
      priv->report_delay = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 report delay: %d ms\n",
	       XCONFIG_GIVEN, priv->report_delay*10);      
      break;
      
    case MAXX:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 maximum x position expected");
      priv->max_x = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 maximum x position: %d\n",
	       XCONFIG_GIVEN, priv->max_x);      
      break;
      
    case MAXY:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 maximum y position expected");
      priv->max_y = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 maximum y position: %d\n",
	       XCONFIG_GIVEN, priv->max_y);      
      break;

    case MIDX:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 maximum x position expected");
      priv->mid_x = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 middle x position: %d\n",
	       XCONFIG_GIVEN, priv->mid_x);      
      break;
      
    case MIDY:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 maximum y position expected");
      priv->mid_y = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 middle y position: %d\n",
	       XCONFIG_GIVEN, priv->mid_y);      
      break;
      
    case MINX:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 minimum x position expected");
      priv->min_x = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 minimum x position: %d\n",
	       XCONFIG_GIVEN, priv->min_x);      
      break;
      
    case MINY:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("B112 minimum y position expected");
      priv->min_y = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 minimum y position: %d\n",
	       XCONFIG_GIVEN, priv->min_y);      
      break;
      
    case B112_EMULATE3:
      priv->emulate3Buttons = TRUE;
      if (xf86Verbose)
	ErrorF("%s B112 3 Button emulation switched on\n");      
      break;
      
    case B112_EM3TIMEOUT:
      if (xf86GetToken(NULL) != NUMBER)
        xf86ConfigError("3 Button emulation time expected");
      priv->emulate3Timeout = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 3 Button emulation time: %d\n",
	       XCONFIG_GIVEN, priv->emulate3Timeout);      
      break;
            
    case DEBUG_LEVEL:
      if (xf86GetToken(NULL) != NUMBER)
	xf86ConfigError("B112 driver debug level expected");
      debug_level = val->num;
      if (xf86Verbose) {
#if DEBUG
	ErrorF("%s B112 debug level sets to %d\n", XCONFIG_GIVEN,
	       debug_level);      
#else
	ErrorF("%s B112 debug not available\n",
	       XCONFIG_GIVEN, debug_level);      
#endif
      }
      break;

    case HISTORY_SIZE:
      if (xf86GetToken(NULL) != NUMBER)
	xf86ConfigError("B112 motion history size expected");
      local->history_size = val->num;
      if (xf86Verbose)
	ErrorF("%s B112 motion history size is %d\n", XCONFIG_GIVEN,
	       local->history_size);      
      break;
	    
    case ALWAYS_CORE:
      xf86AlwaysCore(local, TRUE);
      if (xf86Verbose)
	ErrorF("%s B112 device will always stays core pointer\n",
	       XCONFIG_GIVEN);
      break;

    case EOF:
      FatalError("Unexpected EOF (missing EndSubSection)");
      break;

    default:
      xf86ConfigError("B112 subsection keyword expected");
      break;
    }
  }

  if (priv->max_x - priv->min_x <= 0) {
    ErrorF("%s B112: swap x mode (minimum x position >= maximum x position)\n",
	   XCONFIG_GIVEN);
  }  
  if (priv->max_y - priv->min_y <= 0) {
    ErrorF("%s B112: swap y mode (minimum y position >= maximum y position)\n",
	   XCONFIG_GIVEN, priv->max_y, priv->min_y);
  }
  
  DBG(2, ErrorF("xf86B112Config port name=%s\n", priv->input_dev))

    return Success;
}
#endif


static Bool
xf86B112WriteAck(int fd, unsigned char w) {
  unsigned char c[100];
  int num=0;
  fd_set set;
  Bool ret = !Success;

  /* empty buffer */
  SYSCALL(read(fd, c, 100));

  SYSCALL(write(fd, &w, 1)); 
  DBG(3, ErrorF("B112WriteAck: wrote %02x\n", w));

  FD_ZERO(&set);
  FD_SET(fd, &set);
  SYSCALL(select(fd + 1,  &set,  0, 0, 0));
  SYSCALL(num = read(fd, c, 100));

  while(num > 0) {     
    num--;
    if(c[num] == 0xfa)
      ret =  Success;
    DBG(3, ErrorF("B112WriteAck: got %02x\n", c[num]));
  }

  return (ret);
}

/*
 ***************************************************************************
 *
 * xf86B112Convert --
 *	Convert extended valuators to x and y suitable for core motion
 *	events. Return True if ok and False if the requested conversion
 *	can't be done for the specified valuators.
 *
 ***************************************************************************
 */
static Bool
xf86B112Convert(LocalDevicePtr	local,
		int		first,
		int		num,
		int		v0,
		int		v1,
		int		v2,
		int		v3,
		int		v4,
		int		v5,
		int		*x,
		int		*y)
{
  B112PrivatePtr	priv = (B112PrivatePtr) local->private;

  if (first != 0 || num != 2)
    return FALSE;


  DBG(3, ErrorF("B112Convert: v0(%d), v1(%d)\n",	v0, v1));

  if(v0 > priv->mid_x) 
    *x = (priv->screen_width / 2 * (v0 - priv->mid_x)) 
      / (priv->max_x - priv->mid_x) + priv->screen_width / 2;
  else
    *x = (priv->screen_width / 2 * (v0 - priv->min_x)) 
      / (priv->mid_x - priv->min_x);

  if(v1 > priv->mid_y) 
    *y = priv->screen_height/2 - (priv->screen_height / 2 * (v1 - priv->mid_y)) 
      / (priv->max_y - priv->mid_y);
  else
    *y = priv->screen_height - (priv->screen_height / 2 * (v1 - priv->min_y)) 
      / (priv->mid_y - priv->min_y);
  
  DBG(3, ErrorF("B112Convert: x(%d), y(%d)\n",	*x, *y));

  return TRUE;
}

/*
 ***************************************************************************
 *
 * xf86B112UnConvert --
 *	UnConvert extended valuators to x and y suitable for core motion
 *	events. Return True if ok and False if the requested conversion
 *	can't be done for the specified valuators.
 *
 ***************************************************************************
 */
static Bool
xf86B112UnConvert(B112PrivatePtr	priv,
		  int		v0,
		  int		v1,
		  int		*x,
		  int		*y)
{
  DBG(3, ErrorF("B112UnConvert: v0(%d), v1(%d)\n",	v0, v1));

  if(v0 < priv->rmid_x) 
     *x = (v0 - priv->screen_width / 2) * (priv->max_x - priv->mid_x)
       / (priv->screen_width / 2) + priv->mid_x;
  else
     *x = v0 * (priv->mid_x - priv->min_x) / (priv->screen_width / 2) 
	+ priv->min_x;

  if(v1 < priv->rmid_y) 
     *y = (v1 + priv->screen_height / 2) * (priv->mid_y - priv->max_y)
     / (priv->screen_height / 2) + priv->mid_y;
 else
     *y = (v1 - priv->screen_height) * (priv->min_y - priv->mid_y) 
	/ (priv->screen_height / 2) + priv->min_y; 
  
  DBG(3, ErrorF("B112UnConvert: x(%d), y(%d)\n",	*x, *y));

  return TRUE;
}


/*
 * Lets create a simple finite-state machine:
 *
 *   state[?][0]: action1
 *   state[?][1]: action2
 *   state[?][2]: next state
 *
 *   action > 0: ButtonPress
 *   action = 0: nothing
 *   action < 0: ButtonRelease
 *
 * Why this stuff ??? Normally you cannot press both mousebuttons together, so
 * the mouse reports both pressed at the same time ...
 */

static signed char stateTab[48][3] = {

/* nothing pressed */
  {  0,  0,  0 },	
  {  0,  0,  8 },	/* 1 right -> delayed right */
  {  0,  0,  0 },       /* 2 nothing */
  {  0,  0,  8 },	/* 3 right -> delayed right */
  {  0,  0, 16 },	/* 4 left -> delayed left */
  {  2,  0, 24 },       /* 5 left & right (middle press) -> middle pressed */
  {  0,  0, 16 },	/* 6 left -> delayed left */
  {  2,  0, 24 },       /* 7 left & right (middle press) -> middle pressed */

/* delayed right */
  {  1, -1,  0 },	/* 8 nothing (right event) -> init */
  {  1,  0, 32 },       /* 9 right (right press) -> right pressed */
  {  1, -1,  0 },	/* 10 nothing (right event) -> init */
  {  1,  0, 32 },       /* 11 right (right press) -> right pressed */
  {  1, -1, 16 },       /* 12 left (right event) -> delayed left */
  {  2,  0, 24 },       /* 13 left & right (middle press) -> middle pressed */
  {  1, -1, 16 },       /* 14 left (right event) -> delayed left */
  {  2,  0, 24 },       /* 15 left & right (middle press) -> middle pressed */

/* delayed left */
  {  3, -3,  0 },	/* 16 nothing (left event) -> init */
  {  3, -3,  8 },       /* 17 right (left event) -> delayed right */
  {  3, -3,  0 },	/* 18 nothing (left event) -> init */
  {  3, -3,  8 },       /* 19 right (left event) -> delayed right */
  {  3,  0, 40 },	/* 20 left (left press) -> pressed left */
  {  2,  0, 24 },	/* 21 left & right (middle press) -> pressed middle */
  {  3,  0, 40 },	/* 22 left (left press) -> pressed left */
  {  2,  0, 24 },	/* 23 left & right (middle press) -> pressed middle */

/* pressed middle */
  { -2,  0,  0 },	/* 24 nothing (middle release) -> init */
  { -2,  0,  0 },	/* 25 right (middle release) -> init */
  { -2,  0,  0 },	/* 26 nothing (middle release) -> init */
  { -2,  0,  0 },	/* 27 right (middle release) -> init */
  { -2,  0,  0 },	/* 28 left (middle release) -> init */
  {  0,  0, 24 },	/* 29 left & right -> pressed middle */
  { -2,  0,  0 },	/* 30 left (middle release) -> init */
  {  0,  0, 24 },	/* 31 left & right -> pressed middle */

/* pressed right */
  { -1,  0,  0 },	/* 32 nothing (right release) -> init */
  {  0,  0, 32 },	/* 33 right -> pressed right */
  { -1,  0,  0 },	/* 34 nothing (right release) -> init */
  {  0,  0, 32 },	/* 35 right -> pressed right */
  { -1,  0, 16 },	/* 36 left (right release) -> delayed left */
  { -1,  2, 24 },	/* 37 left & right (r rel, m prs) -> middle pressed */
  { -1,  0, 16 },	/* 38 left (right release) -> delayed left */
  { -1,  2, 24 },	/* 39 left & right (r rel, m prs) -> middle pressed */

/* pressed left */
  { -3,  0,  0 },	/* 40 nothing (left release) -> init */
  { -3,  0,  8 },	/* 41 right (left release) -> delayed right */
  { -3,  0,  0 },	/* 42 nothing (left release) -> init */
  { -3,  0,  8 },	/* 43 right (left release) -> delayed right */
  {  0,  0, 40 },	/* 44 left -> left pressed */
  { -3,  2, 24 },	/* 45 left & right (l rel, mprs) -> middle pressed */
  {  0,  0, 40 },	/* 46 left -> left pressed */
  { -3,  2, 24 },	/* 47 left & right (l rel, mprs) -> middle pressed */
};

/*
 * [JCH-96/01/21] The ALPS GlidePoint pad extends the MS protocol
 * with a fourth button activated by tapping the PAD.
 * The 2nd line corresponds to 4th button on; the drv sends
 * the buttons in the following map (MSBit described first) :
 * 0 | 4th | 1st | 2nd | 3rd
 * And we remap them (MSBit described first) :
 * 0 | 4th | 3rd | 2nd | 1st
 */
static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
			       8, 12, 10, 14,  9, 13, 11, 15,
			      16, 20, 18, 22, 17, 21, 19, 23,
			      24, 28, 26, 30, 25, 29, 27, 31};


void
xf86PostB112ButtonEvent(local, buttons)
    LocalDevicePtr local;
    int buttons;
{
  static OsTimerPtr timer = NULL;
  int         id, change;
  int         truebuttons;
  B112PrivatePtr priv = (B112PrivatePtr)(((LocalDevicePtr)local)->private);

  truebuttons = buttons;
  buttons = reverseMap[buttons];
  
  if (priv->emulate3Buttons) {
    /*
     * Hack to operate the middle button even with Emulate3Buttons set.
     * Modifying the state table to keep track of the middle button state
     * would nearly double its size, so I'll stick with this fix.  - TJW
     */
    change = buttons ^ reverseMap[priv->lastButtons];
    if (change & 02)  {
      xf86PostButtonEvent(local->dev, 0, 2, (buttons & 02), 0, 0);	  
    }
    
    /*
     * emulate the third button by the other two
     */
    if ((id = stateTab[buttons + priv->emulateState][0]) != 0)	{
      xf86PostButtonEvent(local->dev, 0, abs(id), (id >= 0), 0, 0);
    }
    
    if ((id = stateTab[buttons + priv->emulateState][1]) != 0) {
      xf86PostButtonEvent(local->dev, 0, abs(id), (id >= 0), 0, 0);
    }
    
    priv->emulateState = stateTab[buttons + priv->emulateState][2];
    if (stateTab[buttons + priv->emulateState][0] ||
	stateTab[buttons + priv->emulateState][1]) {
      priv->truebuttons = truebuttons;
      timer = TimerSet(timer, 0, priv->emulate3Timeout, buttonTimer,
		       (pointer)local);
    }
    else {
      if (timer) {
	TimerFree(timer);
	timer = NULL;
      }
    }
  }
  else {
    /*
     * real three button event
     * Note that xf86Info.lastButtons has the hardware button mapping which
     * is the reverse of the button mapping reported to the server.
     */
    change = buttons ^ reverseMap[priv->lastButtons];
    while (change) {
      id = ffs(change);
      change &= ~(1 << (id-1));
      xf86PostButtonEvent(local->dev, 0, id, (buttons&(1<<(id-1))), 0, 0);
    }
  }
  priv->lastButtons = truebuttons;
}

/*
 ***************************************************************************
 *
 * buttonTimer --
 *	Sends a delayed Button event
 *
 ***************************************************************************
 */
static CARD32
buttonTimer(timer, now, local)
     OsTimerPtr timer;
     CARD32 now;
     pointer local;
{
  B112PrivatePtr priv = (B112PrivatePtr)(((LocalDevicePtr)local)->private);

  xf86PostB112ButtonEvent((LocalDevicePtr)local, priv->truebuttons);
  return(0);
}

static void
xf86B112SetRelative(LocalDevicePtr	local)
{
   /* enable */
   xf86B112WriteAck(local->fd, 0xe6);
   xf86B112WriteAck(local->fd, 0xf3);
   xf86B112WriteAck(local->fd, 0x28);
   xf86B112WriteAck(local->fd, 0xe8);
   xf86B112WriteAck(local->fd, 0x3);
   xf86B112WriteAck(local->fd, 0xe8);
   xf86B112WriteAck(local->fd, 0x7);
   xf86B112WriteAck(local->fd, 0xf4);
   xf86B112WriteAck(local->fd, 0xf4);
   xf86B112WriteAck(local->fd, 0xf4);
}

static void
xf86B112Reset(LocalDevicePtr	local)
{
   char rep[2];

   xf86B112WriteAck(local->fd, 0xf5);
   /* reset */
   xf86B112WriteAck(local->fd, 0xff);
   SYSCALL(fcntl(local->fd, F_SETFL, O_RDWR));
   read(local->fd, rep, 2);
   SYSCALL(fcntl(local->fd, F_SETFL, O_RDWR|O_NDELAY));
   xf86B112WriteAck(local->fd, 0xf4);
   xf86B112SetRelative(local);	
}

/*
 ***************************************************************************
 *
 * xf86B112ReadInput --
 *	Read all pending report packets from the touchscreen and enqueue
 *	them.
 *	If a packet is not fully received it is deferred until the next
 *	call to the function.
 *
 ***************************************************************************
 */
static void
xf86B112ReadInput(LocalDevicePtr	local)
{
  B112PrivatePtr		priv = (B112PrivatePtr)(local->private);
  int				cur_x, cur_y, dx, dy;
  int				buttons;
  unsigned char 		*bptr;
  struct vt_stat vts;
  static Bool vt_switched = FALSE;

  /* Test if out VT is active */
  if (ioctl(xf86Info.consoleFd, VT_GETSTATE, &vts) == 0) {
    if (vts.v_active != xf86Info.vtno) {
      /* let others read the data */
      usleep(100 * 1000);
      vt_switched = TRUE;
      return;
    }
  }
  
  DBG(4, ErrorF("Entering ReadInput\n"));


  if(B112_BUFSIZE > priv->packet_buf_p) {
    int num_bytes=0;

    SYSCALL(num_bytes = read(local->fd, &priv->packet_buf[priv->packet_buf_p], 
			     B112_BUFSIZE - priv->packet_buf_p));

    if(vt_switched == TRUE) {
      priv->packet_buf_p = 0;
      vt_switched = FALSE;
      return;
    }

    if(num_bytes >= 1) {
      priv->packet_buf_p += num_bytes; 
    }

    if (priv->packet_buf_p < B112_PACKET_SIZE) 
      return;
  }
  
  bptr = priv->packet_buf - B112_PACKET_SIZE;
  
  while(priv->packet_buf_p >= B112_PACKET_SIZE) {
     bptr += B112_PACKET_SIZE;  
     if (((bptr[0] & 0xC8) == 8)
	 || ((bptr[0] & 0xa) == 0 && (bptr[1] != 0 || bptr[2] != 0))
	 || (bptr[0] == 0 && bptr[1] == 0 && bptr[2] == 0)) {
	priv->packet_buf_p -= B112_PACKET_SIZE;
     } else {
	DBG(0, ErrorF("Reseting ReadInput %02x %02x %02x\n", 
		      (int)bptr[0], 
		      (int)bptr[1],
		      (int)bptr[2]));
	xf86B112WriteAck(local->fd, 0xf5);
	/* reset */
	xf86B112Reset(local);	
	priv->packet_buf_p = 0;
	continue;
     }

    
    /*
     * Process B112_TOUCHs here.
     */
  
  
    if (bptr[0] & 8) {

      /*
       * Send events from the pointer.
       */

      buttons = (bptr[0] & 0x02) >> 1 |       /* Right */
	        (bptr[0] & 0x01) << 2;        /* Left */

      if(bptr[1] != 0)
	dx=  (bptr[0] & 0x10) ? bptr[1]-256 : bptr[1];
      else
	dx = 0;

      if(bptr[2] != 0)
	dy= ((bptr[0] & 0x20) ? bptr[2]-256 : bptr[2]);
      else
	dy = 0;

      /* post an event */
      DBG(4, ErrorF("Post mouse packet\n"));

      priv->cur_x += dx;
      priv->cur_x = ( priv->cur_x > priv->max_x) ? priv->max_x : priv->cur_x;
      priv->cur_x = ( priv->cur_x < priv->min_x) ? priv->min_x : priv->cur_x;
      
      priv->cur_y += dy;
      priv->cur_y = ( priv->cur_y > priv->max_y) ? priv->max_y : priv->cur_y;
      priv->cur_y = ( priv->cur_y < priv->min_y) ? priv->min_y : priv->cur_y;
      
//	 xf86PostMotionEvent(local->dev, FALSE, 0, 2, cdx, cdy);

      xf86PostMotionEvent(local->dev, TRUE, 0, 2, priv->cur_x, priv->cur_y);

      xf86PostB112ButtonEvent(local, buttons);

      continue;
    }

    GET_TIME(priv->lasttv);

    /*
     * Send events from the stylus.
     *
     * We *must* generate a motion before a button change if pointer
     * location has changed as DIX assumes this. This is why we always
     * emit a motion, regardless of the kind of packet processed.
     */

    if (!(bptr[0]&0x04)) {

      /*
       * Emit a button release.
       */
      priv->was_pressed &= ~1;

      DBG(3, ErrorF("Stylus: %s\n", ((bptr[0] & 4)) ? "Press" : "Release"));

      /*
       * Ignore position, take last one.
       */

      xf86PostMotionEvent(local->dev, TRUE, 0, 2, priv->cur_x, priv->cur_y);
      xf86PostButtonEvent(local->dev, TRUE, 1, 0, 0, 0);

      continue;
    }

  
    priv->cur_x = cur_x =  (((unsigned int)bptr[0] & 0x30) << 4) + bptr[1];
    priv->cur_y = cur_y = ((((unsigned int)bptr[0] & 0xc0) << 2) + bptr[2]);

    DBG(3, ErrorF("TouchScreen: x(%d), y(%d)\n", priv->cur_x, priv->cur_y));
  
    xf86PostMotionEvent(local->dev, TRUE, 0, 2, priv->cur_x, priv->cur_y);
  
    if(!(priv->was_pressed & 1)) {
      /*
       * Emit a button press.
       */
      priv->was_pressed |= 1;
      xf86PostButtonEvent(local->dev, TRUE, 1, 1, 0, 0);
      DBG(3, ErrorF("Stylus: %s \n", ((bptr[0] & 4)) ? "Press" : "Release"));
    }
  }

  if(unlink(B112_RESET) == 0) {
     /* reset */
     DBG(0, ErrorF("Reseting TouchScreen\n")); 
     xf86B112Reset(local);
  }

}

/*
 ***************************************************************************
 *
 * xf86B112PrintIdent --
 *	Print type of touchscreen and features on controller board.
 *
 ***************************************************************************
 */
static void
xf86B112PrintIdent(unsigned char	*packet,
		   B112PrivatePtr	priv)
{
  ErrorF("%s B112 touchscreen", XCONFIG_PROBED);
}



/*
 ***************************************************************************
 *
 * xf86B112Control --
 *
 ***************************************************************************
 */
static Bool
xf86B112Control(DeviceIntPtr	dev,
		int		mode)
{
  LocalDevicePtr	local = (LocalDevicePtr) dev->public.devicePrivate;
  B112PrivatePtr		priv = (B112PrivatePtr)(local->private);
  unsigned char		map[] = { 0, 1, 2, 3 };

  switch(mode) {

  case DEVICE_INIT:
    {  
      DBG(2, ErrorF("B112 touchscreen init...\n"));
      
      if (priv->screen_no >= screenInfo.numScreens ||
	  priv->screen_no < 0)
	priv->screen_no = 0;
      priv->screen_width = screenInfo.screens[priv->screen_no]->width;
      priv->screen_height = screenInfo.screens[priv->screen_no]->height;
      priv->cur_x = priv->mid_x;
      priv->cur_y = priv->mid_y;

      xf86B112Convert(local, 0, 2, priv->mid_x, priv->mid_y, 0, 0, 0, 0, 
		      &priv->rmid_x, &priv->rmid_y);


      /*
       * Device reports button press for up to 3 buttons.
       */
      if (InitButtonClassDeviceStruct(dev, 3, map) == FALSE) {
	ErrorF("Unable to allocate B112 touchscreen ButtonClassDeviceStruct\n");
	return !Success;
      }
      
      /*
       * Device reports motions on 2 axes in absolute coordinates.
       * Axes min and max values are reported in raw coordinates.
       * Resolution is computed roughly by the difference between
       * max and min values scaled from the approximate size of the
       * screen to fit one meter.
       */
      if (InitValuatorClassDeviceStruct(dev, 2, xf86GetMotionEvents,
					local->history_size, Absolute) == FALSE) {
	ErrorF("Unable to allocate B112 touchscreen ValuatorClassDeviceStruct\n");
	return !Success;
      }
      else {
	InitValuatorAxisStruct(dev, 0, priv->min_x, priv->max_x,
			       1024,
			       0     /* min_res */,
			       1024  /* max_res */);
	InitValuatorAxisStruct(dev, 1, priv->min_y, priv->max_y,
			       1024,
			       0     /* min_res */,
			       1024 /* max_res */);
      }

      if (InitFocusClassDeviceStruct(dev) == FALSE) {
	ErrorF("Unable to allocate B112 touchscreen FocusClassDeviceStruct\n");
      }
      
      /*
       * Allocate the motion events buffer.
       */
      xf86MotionHistoryAllocate(local);
      
      /*
       * This once has caused the server to crash after doing an xalloc & strcpy ??
       */
      AssignTypeAndName(dev, local->atom, local->name);
      
      DBG(2, ErrorF("Done.\n"));
      return Success;
    
    case DEVICE_ON:
      DBG(2, ErrorF("B112 touchscreen on...\n"));

      if (local->fd < 0) {
	DBG(2, ErrorF("B112 touchscreen opening : %s\n",
		      priv->input_dev));
	SYSCALL(local->fd = open(priv->input_dev, O_RDWR|O_NDELAY, 0));
	if (local->fd < 0) {
	  Error("Unable to open B112 touchscreen device");
	  return !Success;
	}      
      }      

      /* disable */
      if(xf86B112WriteAck(local->fd, 0xf5) != Success)
	return !Success;
      /* reset */
      if(xf86B112WriteAck(local->fd, 0xff) != Success)
	return !Success;
      else {
	char rep[2];
	SYSCALL(fcntl(local->fd, F_SETFL, O_RDWR));
	read(local->fd, rep, 2);
	SYSCALL(fcntl(local->fd, F_SETFL, O_RDWR|O_NDELAY));
      }
      /* enable */
      xf86B112SetRelative(local);	
      /* enable */
      if(xf86B112WriteAck(local->fd, 0xf4) != Success)
	return !Success;


      SYSCALL(fcntl(local->fd, F_SETFL, O_RDONLY|O_NDELAY));

      AddEnabledDevice(local->fd);
      dev->public.on = TRUE;  
    }
    
    DBG(2, ErrorF("Done\n"));
    return Success;
    
    /*
     * Deactivate the device. After this, the device will not emit
     * events until a subsequent DEVICE_ON. Thus, we can momentarily
     * close the port.
     */
  case DEVICE_OFF:
    DBG(2, ErrorF("B112 touchscreen off...\n"));
    dev->public.on = FALSE;
    if (local->fd >= 0)
      RemoveEnabledDevice(local->fd);
    SYSCALL(close(local->fd));
    local->fd = -1;
    DBG(2, ErrorF("Done\n"));
    return Success;
    
    /*
     * Final close before server exit. This is used during server shutdown.
     * Close the port and free all the resources.
     */
  case DEVICE_CLOSE:
    DBG(2, ErrorF("B112 touchscreen close...\n"));
    dev->public.on = FALSE;
    if (local->fd >= 0)
      RemoveEnabledDevice(local->fd);
    SYSCALL(close(local->fd));
    local->fd = -1;
    DBG(2, ErrorF("Done\n"));
    return Success;

  default:
    ErrorF("unsupported mode=%d\n", mode);
    return !Success;
  }
}

/*
 ***************************************************************************
 *
 * xf86B112Allocate --
 *
 ***************************************************************************
 */
static LocalDevicePtr
xf86B112Allocate(
#if NeedFunctionPrototypes
		 void
#endif
		 )
{
  LocalDevicePtr        local = (LocalDevicePtr) xalloc(sizeof(LocalDeviceRec));
  B112PrivatePtr         priv = (B112PrivatePtr) xalloc(sizeof(B112PrivateRec));
  
  priv->input_dev = B112_PORT;
  priv->min_x = DEFAULT_MIN_X;
  priv->mid_x = DEFAULT_MID_X;
  priv->max_x = DEFAULT_MAX_X;
  priv->min_y = DEFAULT_MIN_Y;
  priv->mid_y = DEFAULT_MID_Y;
  priv->max_y = DEFAULT_MAX_Y;
  priv->untouch_delay = 0;
  priv->screen_no = 0;
  priv->screen_width = -1;
  priv->screen_height = -1;
  priv->inited = 0;
  priv->packet_buf_p = 0;
  priv->was_pressed = 0;
  priv->avgx = -1;
  priv->emulate3Buttons = FALSE;
  priv->emulate3Timeout = 50;
  GET_TIME(priv->lasttv);
  priv->lasttv.tv_sec += 5;

  local->name = XI_TOUCHSCREEN;
  local->flags = XI86_NO_OPEN_ON_INIT;
#if !defined(sun) || defined(i386)
  local->device_config = xf86B112Config;
#endif
  local->device_control = xf86B112Control;
  local->read_input   = xf86B112ReadInput;
  local->control_proc = NULL;
  local->close_proc   = NULL;
  local->switch_mode  = NULL;
  local->conversion_proc = xf86B112Convert;
  local->reverse_conversion_proc = NULL;
  local->fd	      = -1;
  local->atom	      = 0;
  local->dev	      = NULL;
  local->private      = priv;
  local->type_name    = "B112 TouchScreen";
  local->history_size = 0;
  
  return local;
}

/*
 ***************************************************************************
 *
 * B112 device association --
 *
 ***************************************************************************
 */
DeviceAssocRec B112_assoc =
{
  "B112",                /* config_section_name */
  xf86B112Allocate               /* device_allocate */
};

#ifdef DYNAMIC_MODULE
/*
 ***************************************************************************
 *
 * entry point of dynamic loading
 *
 ***************************************************************************
 */
int
#ifndef DLSYM_BUG
init_module(unsigned long	server_version)
#else
     init_xf86B112(unsigned long      server_version)
#endif
{
  xf86AddDeviceAssoc(&B112_assoc);

  if (server_version != XF86_VERSION_CURRENT) {
    ErrorF("Warning: B112 module compiled for version%s\n", XF86_VERSION);
    return 0;
  } else {
    return 1;
  }
}
#endif

#ifdef XFree86LOADER
/*
 * Entry point for the loader code
 */
XF86ModuleVersionInfo xf86B112Version = {
  "xf86B112",
  MODULEVENDORSTRING,
  MODINFOSTRING1,
  MODINFOSTRING2,
  XF86_VERSION_CURRENT,
  1, 0, 0,
  ABI_CLASS_XINPUT,
  ABI_XINPUT_VERSION,
  MOD_CLASS_XINPUT,
  { 0, 0, 0, 0 }
};

void
xf86B112ModuleInit(data, magic)
     pointer *data;
     INT32 *magic;
{
  static int cnt = 0;

  switch (cnt) {
  case 0:
    *magic = MAGIC_VERSION;
    *data = &xf86B112Version;
    cnt++;
    break;
      
  case 1:
    *magic = MAGIC_ADD_XINPUT_DEVICE;
    *data = &B112_assoc;
    cnt++;
    break;

  default:
    *magic = MAGIC_DONE;
    *data = NULL;
    break;
  } 
}
#endif

